
<p>
Весь функционал в рабочем состоянии и готов для загрузки статей и книг. Пока мы собрали мало текстов, но наш зверь готов сожрать много больше. Если вы хотите поделиться с нами своими любимыми текстами, просто кликайте по кнопке «Добавить текст» и оформляйте запись согласно Инструкции - не нужно регистрироваться и создавать аккаунт. Правда, пока придется несколько подождать, прежде чем ваш текст откроется миру.
</p>

<div class="center">

<p>
This is centered
Centered again
hello
</p>

</div>

<div class="center">

<p>
centered
</p>

</div>

<blockquote>

<p>
This is a quotation
and here it continues
</p>

</blockquote>

<blockquote>

<p>
And here too.
</p>

</blockquote>
<div class="verse">
this is verse<br />
&nbsp;&nbsp;and this is the same<br />
same here<br />
hello
<br /><br />
hello hello
</div>

<p>
 this is a tab
</p>
<h2 id="toc1">First level and</h2>

<h3 id="toc2">Second level</h3>

<h4 id="toc3">Third level</h4>

<h5 id="toc4">Fourth level</h5>

<h6>Fifth level
and here</h6>


<ul>
<li>
<p>
bullet item 1, line 1
</p>

<p>
bullet item 1, line 2
</p>

<ol>
<li>
<p>
Enum line 1
</p>

<p>
Enum line 2
</p>

<p>
Term1 ::
This is a first definition
And it has two lines;
no, make that three.
</p>

</li>

</ol>

</li>
<li>
<p>
bullet item 2, line 1
</p>

<p>
bullet item 2, line 2
</p>

<p>
Term1 ::
This is a first definition
And it has two lines;
no, make that three.
</p>

</li>

</ul>

<p>
Normal text.
</p>

<ul>
<li>
<p>
bullet item one
</p>

</li>
<li>
<p>
bullet item two
</p>

</li>

</ul>

<p>
Normal text.
</p>

<ol>
<li>
<p>
Enum item one
</p>

</li>
<li>
<p>
Enum item two
</p>

</li>

</ol>

<p>
Term1 ::
This is a first definition
And it has two lines;
no, make that three.
</p>

<p>
Term2 :: This is a second definition
</p>

<table>
 <thead>
  <tr>
   <th>
     Double bars  
   </th>
   <th>
 Separate header fields
   </th>
  </tr>
 </thead>
 <tfoot>
  <tr>
   <td>
     Triple bars 
   </td>
   <td>
 Separate footer fields
   </td>
  </tr>
 </tfoot>
 <tbody>
  <tr>
   <td>
     Single bars   
   </td>
   <td>
 Separate body fields
   </td>
  </tr>
  <tr>
   <td>
     Here are more 
   </td>
   <td>
 body fields
   </td>
  </tr>
 </tbody>
</table>

<p>
Regular again.
</p>

<!-- start comment -->
<div class="comment"><span class="commentmarker">{{COMMENT:</span> 
; this is a comment
<span class="commentmarker">END_COMMENT}}:</span>
</div>
<!-- stop comment -->

<p>
 ; while this is not
</p>
